/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Maximilian.Schnarr@googlemail.com>
 *
 */

#include <gdk/gdk.h>

#include <livewallpaper/core.h>
#include <libpeas/peas.h>

#include "particle.h"
#include "noise.h"


#define NOISE_PLUGIN_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NOISE_TYPE_PLUGIN, NoisePluginPrivate))

struct _NoisePluginPrivate
{
	GSettings *settings;

	NoiseParticleSystem *ps;

	LwBackground *background;
};

static void lw_livewallpaper_plugin_iface_init(LwLivewallpaperPluginInterface *iface);

G_DEFINE_DYNAMIC_TYPE_EXTENDED(NoisePlugin, noise_plugin, PEAS_TYPE_EXTENSION_BASE, 0,
							   G_IMPLEMENT_INTERFACE_DYNAMIC(LW_TYPE_LIVEWALLPAPER_PLUGIN,
							   								 lw_livewallpaper_plugin_iface_init))

static void
noise_plugin_adjust_viewport(LwLivewallpaperPlugin *plugin, LwOutput *output)
{
	guint width, height;
	(void) plugin;

	/* Enable blending */
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

	/* Save and change texture environment */
	glPushAttrib(GL_TEXTURE_BIT);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* Save and reset the coordinate system before touching */
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();

	width = lw_output_get_width(output);
	height = lw_output_get_height(output);

	if(width > height)
		glOrtho(0.0, 1.0, 0.0, ((gdouble) height) / ((gdouble) width), 0.0, 1.0);
	else
		glOrtho(0.0, ((gdouble) width) / ((gdouble) height), 0.0, 1.0, 0.0, 1.0);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	glEnable(GL_VERTEX_PROGRAM_POINT_SIZE);
}

static void
noise_plugin_prepare_paint(LwLivewallpaperPlugin *plugin, gint ms_since_last_paint)
{
	NoisePlugin *self = NOISE_PLUGIN(plugin);

	noise_particle_system_update(self->priv->ps, ms_since_last_paint);
}

static void
noise_plugin_paint(LwLivewallpaperPlugin *plugin, LwOutput *output)
{
	NoisePlugin *self = NOISE_PLUGIN(plugin);

	/* Clear color buffer and draw background image */
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	lw_background_draw(self->priv->background, output);

	noise_particle_system_draw(self->priv->ps);
}

static void
noise_plugin_restore_viewport(LwLivewallpaperPlugin *plugin)
{
	(void) plugin;

	glDisable(GL_VERTEX_PROGRAM_POINT_SIZE);

	/* Restore the coordinate system */
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	/* Restore the original perspective */
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	/* Restore texture environment */
	glPopAttrib();

	/* Restore blending */
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_BLEND);
}

static void
noise_plugin_init(NoisePlugin *self)
{
	self->priv = NOISE_PLUGIN_GET_PRIVATE(self);

	self->priv->ps = g_object_new(NOISE_TYPE_PARTICLE_SYSTEM, NULL);

	lw_noise_init();
}

static void
noise_plugin_init_plugin(LwLivewallpaperPlugin *plugin)
{
	NoisePlugin *self = NOISE_PLUGIN(plugin);
	GSettings *settings = g_settings_new("net.launchpad.livewallpaper.plugins.noise");
	gchar *filename, *datadir = peas_extension_base_get_data_dir(PEAS_EXTENSION_BASE(self));
	GError *error = NULL;

	noise_particle_system_set_data_dir(self->priv->ps, datadir);

	/* Load background image */
	filename = g_build_filename(datadir, "background.png", NULL);
	self->priv->background = lw_background_new_with_default_image(filename, LwBackgroundZoom, &error);
	if(error != NULL)
	{
		g_warning("Could not load default background texture: %s", error->message);
		g_error_free(error);
		error = NULL;
	}
	g_free(filename);

	/* Bind settings to properties */
	g_settings_bind(settings, "particle-count", self->priv->ps, "particle-count", G_SETTINGS_BIND_GET);
	g_settings_bind(settings, "fading-time", self->priv->ps, "fading-time", G_SETTINGS_BIND_GET);

	self->priv->settings = settings;
}

static void
noise_plugin_dispose(GObject *object)
{
	NoisePlugin *self = NOISE_PLUGIN(object);

	if(self->priv->settings)
	{
		g_object_unref(self->priv->settings);
		self->priv->settings = NULL;
	}

	if(self->priv->ps)
	{
		g_object_unref(self->priv->ps);
		self->priv->ps = NULL;
	}

	if(self->priv->background)
	{
		g_object_unref(self->priv->background);
		self->priv->background = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(noise_plugin_parent_class)->dispose(object);
}

static void
noise_plugin_class_init(NoisePluginClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->dispose = noise_plugin_dispose;

	g_type_class_add_private(klass, sizeof(NoisePluginPrivate));
}

static void
noise_plugin_class_finalize(NoisePluginClass *klass)
{
	(void) klass;
}

static void
lw_livewallpaper_plugin_iface_init(LwLivewallpaperPluginInterface *iface)
{
	iface->init_plugin = noise_plugin_init_plugin;

	iface->adjust_viewport = noise_plugin_adjust_viewport;
	iface->prepare_paint = noise_plugin_prepare_paint;
	iface->paint = noise_plugin_paint;
	iface->restore_viewport = noise_plugin_restore_viewport;
}

G_MODULE_EXPORT void
peas_register_types(PeasObjectModule *module)
{
	noise_plugin_register_type(G_TYPE_MODULE(module));
	peas_object_module_register_extension_type(module, LW_TYPE_LIVEWALLPAPER_PLUGIN, NOISE_TYPE_PLUGIN);
}

