/*
 *
 * Livewallpaper
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Copyright (C) 2012 Maximilian Schnarr <Maximilian.Schnarr@googlemail.com>
 *
 */

#include <math.h>
#include <glib-object.h>
#include <gdk/gdk.h>
#include <livewallpaper/core.h>

#include "particle.h"


typedef struct
{
	/* Position of the particle */
	gfloat x, y;

	/* */
	gfloat speed;

	/* */
	gfloat size;

	/* */
	gfloat wander;

	/* */
	int lifetime;

	/* */
	int alive;

	/* */
	gfloat alpha;
} Particle;

#define _2PI (gfloat) 6.283185

#define NOISE_PARTICLE_SYSTEM_GET_PRIVATE(obj)  (G_TYPE_INSTANCE_GET_PRIVATE ((obj), NOISE_TYPE_PARTICLE_SYSTEM, NoiseParticleSystemPrivate))

struct _NoiseParticleSystemPrivate
{
	guint particle_count;
	gint fading_time;

	gchar *datadir;

	GArray *particles;
	gfloat *vertices;
	gfloat *alpha;
	gfloat *size;

	GLuint vertex_buffer;
	GLuint alpha_buffer;
	GLuint size_buffer;

	LwProgram *prog;
	LwTexture *texture;
};

enum
{
    PROP_0,

    PROP_PARTICLE_COUNT,
    PROP_FADING_TIME,

    N_PROPERTIES
};

static GParamSpec *obj_properties[N_PROPERTIES] = {NULL, };

G_DEFINE_TYPE(NoiseParticleSystem, noise_particle_system, G_TYPE_OBJECT)


static void
noise_particle_system_set_property(GObject *object,
                                   guint property_id,
                                   const GValue *value,
                                   GParamSpec *pspec)
{
	NoiseParticleSystem *self = NOISE_PARTICLE_SYSTEM(object);

	switch(property_id)
	{
		case PROP_PARTICLE_COUNT:
			noise_particle_system_set_particle_count(self, g_value_get_uint(value));
			break;

		case PROP_FADING_TIME:
			noise_particle_system_set_fading_time(self, g_value_get_uint(value));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static void
noise_particle_system_get_property(GObject *object,
                                   guint property_id,
                                   GValue *value,
                                   GParamSpec *pspec)
{
	NoiseParticleSystem *self = NOISE_PARTICLE_SYSTEM(object);

	switch(property_id)
	{
		case PROP_PARTICLE_COUNT:
			g_value_set_uint(value, noise_particle_system_get_particle_count(self));
			break;

		case PROP_FADING_TIME:
			g_value_set_uint(value, noise_particle_system_get_fading_time(self));
			break;

		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
			break;
	}
}

static inline void
noise_particle_system_init_particle(NoiseParticleSystem *self, Particle *particle)
{
	particle->x = randf();
	particle->y = randf();

	particle->speed = rand2f(0.000005f, 0.0001f);
	particle->size = rand2f(10.0f, 50.0f);
	particle->wander = rand2f(0.5f, 1.5f);

	particle->lifetime = rand2f(30000, 80000) + 2 * self->priv->fading_time;
	particle->alive = 0;
	particle->alpha = rand2f(0.1f, 1.0f);
}

void
noise_particle_system_set_particle_count(NoiseParticleSystem *self, guint count)
{
	g_array_set_size(self->priv->particles, count);

	/* Append particles */
	if(self->priv->particle_count < count)
	{
		guint i;

		for(i = self->priv->particle_count; i < count; i++)
		{
			Particle *particle = &g_array_index(self->priv->particles, Particle, i);
			noise_particle_system_init_particle(self, particle);
		}
	}

	self->priv->vertices = g_realloc(self->priv->vertices, 2 * count * sizeof(gfloat));
	self->priv->alpha = g_realloc(self->priv->alpha, count * sizeof(gfloat));
	self->priv->size = g_realloc(self->priv->size, count * sizeof(gfloat));
	self->priv->particle_count = count;

	/* Resize OpenGL buffers */
	glBindBuffer(GL_ARRAY_BUFFER, self->priv->vertex_buffer);
	glBufferData(GL_ARRAY_BUFFER, 2 * self->priv->particle_count * sizeof(gfloat), self->priv->vertices, GL_STREAM_DRAW);

	glBindBuffer(GL_ARRAY_BUFFER, self->priv->alpha_buffer);
	glBufferData(GL_ARRAY_BUFFER, self->priv->particle_count * sizeof(gfloat), self->priv->alpha, GL_STREAM_DRAW);

	glBindBuffer(GL_ARRAY_BUFFER, self->priv->size_buffer);
	glBufferData(GL_ARRAY_BUFFER, self->priv->particle_count * sizeof(gfloat), self->priv->size, GL_STREAM_DRAW);	

	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

guint
noise_particle_system_get_particle_count(NoiseParticleSystem *self)
{
	return self->priv->particle_count;
}

void
noise_particle_system_set_fading_time(NoiseParticleSystem *self, guint fading_time)
{
	self->priv->fading_time = fading_time;
}

guint
noise_particle_system_get_fading_time(NoiseParticleSystem *self)
{
	return self->priv->fading_time;
}

void
noise_particle_system_set_data_dir(NoiseParticleSystem *self, gchar *datadir)
{
	self->priv->datadir = g_strdup(datadir);

	if(self->priv->datadir)
	{
		LwShader *shader;
		gchar *filename;
		GError *error = NULL;

		/* (Re)load particle texture */
		if(self->priv->texture)
		{
			g_object_unref(self->priv->texture);
			self->priv->texture = NULL;
		}

		filename = g_build_filename(self->priv->datadir, "particle.png", NULL);
		self->priv->texture = lw_texture_new_from_file(filename, &error);
	    if(error != NULL)
	    {
		    g_critical("Could not load particle texture: %s", error->message);
		    g_error_free(error);
    		error = NULL;
    	}
    	g_free(filename);

    	/* (Re)load program */
    	if(self->priv->prog)
    	{
    		g_object_unref(self->priv->prog);
    		self->priv->prog = NULL;
    	}

	    self->priv->prog = g_object_new(LW_TYPE_PROGRAM, NULL);

		/* Vertex shader */
		filename = g_build_filename(self->priv->datadir, "vert.glsl", NULL);
	    shader = lw_shader_new_from_file(filename, GL_VERTEX_SHADER, &error);
	    if(error != NULL)
	    {
		    g_critical("Could not load vertex shader: %s", error->message);
		    g_error_free(error);
	    	error = NULL;
	    }
	    else
	    {
	    	lw_shader_compile(shader);
	    	lw_program_attach_shader(self->priv->prog, shader);
	    }
	    g_object_unref(shader);
	    g_free(filename);

	    /* Fragment shader */
		filename = g_build_filename(self->priv->datadir, "frag.glsl", NULL);
	    shader = lw_shader_new_from_file(filename, GL_FRAGMENT_SHADER, &error);
	    if(error != NULL)
	    {
		    g_critical("Could not load fragment shader: %s", error->message);
		    g_error_free(error);
	    	error = NULL;
	    }
	    else
	    {
	    	lw_shader_compile(shader);
	    	lw_program_attach_shader(self->priv->prog, shader);
	    }
	    g_object_unref(shader);
	    g_free(filename);

	    lw_program_link(self->priv->prog);
	}
}

void
noise_particle_system_update(NoiseParticleSystem *self, gint ms_since_last_paint)
{
	gfloat *v = self->priv->vertices;
	guint i;

	for(i = 0; i < self->priv->particle_count; i++)
	{
		float noise, angle, speed;
    	Particle *p = &g_array_index(self->priv->particles, Particle, i);

		p->alive += ms_since_last_paint;
		if(p->lifetime < p->alive || p->x < -0.1f || p->x > 1.1f || p->y < -0.1f || p->y > 1.1f)
		{
			noise_particle_system_init_particle(self, p);
		}

		/* Update position */
    	noise = lw_simplex_noise_2f(p->x, p->y);
		angle = _2PI * noise * p->wander;
		speed = p->speed * ms_since_last_paint * noise + 0.00005f;

		p->x += cos(angle) * speed * 0.33;
		p->y += sin(angle) * speed * 0.33;

		v[0] = p->x;
		v[1] = p->y;
		v += 2;

		/* Update alpha */
		if(p->alive < self->priv->fading_time)
		{
			/* Fade in */
			self->priv->alpha[i] = p->alpha * p->alive / self->priv->fading_time;
		}
		else if(p->lifetime - p->alive < self->priv->fading_time)
		{
			/* Fade out */
			self->priv->alpha[i] = p->alpha * (p->lifetime - p->alive) / self->priv->fading_time;
		}
		else
		{
			self->priv->alpha[i] = p->alpha;
		}

		/* Update size */
		self->priv->size[i] = p->size;
	}

	/* Put data into the buffers */
	glBindBuffer(GL_ARRAY_BUFFER, self->priv->vertex_buffer);
	glBufferSubData(GL_ARRAY_BUFFER, 0, 2 * self->priv->particle_count * sizeof(gfloat), self->priv->vertices);

	glBindBuffer(GL_ARRAY_BUFFER, self->priv->alpha_buffer);
	glBufferSubData(GL_ARRAY_BUFFER, 0, self->priv->particle_count * sizeof(gfloat), self->priv->alpha);

	glBindBuffer(GL_ARRAY_BUFFER, self->priv->size_buffer);
	glBufferSubData(GL_ARRAY_BUFFER, 0, self->priv->particle_count * sizeof(gfloat), self->priv->size);

	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void
noise_particle_system_draw(NoiseParticleSystem *self)
{
	int positionIndex, alphaIndex, sizeIndex, texIndex;

	if(!self->priv->prog) return;
	if(self->priv->texture) lw_texture_enable(self->priv->texture);

	glEnable(GL_POINT_SPRITE);

	/* Save texture environment */
	glPushAttrib(GL_TEXTURE_BIT);

	glPointParameterf(GL_POINT_FADE_THRESHOLD_SIZE, 60.0f);
	glPointParameterf(GL_POINT_SIZE_MIN, 0.1f);

	glTexEnvi(GL_POINT_SPRITE, GL_COORD_REPLACE, GL_TRUE);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	/* Render particles using buffers */
	lw_program_enable(self->priv->prog);

	positionIndex = lw_program_get_attrib_location(self->priv->prog, "position");
	glEnableVertexAttribArray(positionIndex);
	glBindBuffer(GL_ARRAY_BUFFER, self->priv->vertex_buffer);
	glVertexAttribPointer(positionIndex, 2, GL_FLOAT, GL_FALSE, 0, 0);

	alphaIndex = lw_program_get_attrib_location(self->priv->prog, "alpha");
	glEnableVertexAttribArray(alphaIndex);
	glBindBuffer(GL_ARRAY_BUFFER, self->priv->alpha_buffer);
	glVertexAttribPointer(alphaIndex, 1, GL_FLOAT, GL_FALSE, 0, 0);

	sizeIndex = lw_program_get_attrib_location(self->priv->prog, "size");
	glEnableVertexAttribArray(sizeIndex);
	glBindBuffer(GL_ARRAY_BUFFER, self->priv->size_buffer);
	glVertexAttribPointer(sizeIndex, 1, GL_FLOAT, GL_FALSE, 0, 0);

	glBindBuffer(GL_ARRAY_BUFFER, 0);

	texIndex = lw_program_get_uniform_location(self->priv->prog, "texture");
	glUniform1i(texIndex, 0);

	glDrawArrays(GL_POINTS, 0, self->priv->particle_count);

	glDisableVertexAttribArray(positionIndex);
	glDisableVertexAttribArray(alphaIndex);
	glDisableVertexAttribArray(sizeIndex);

	lw_program_disable(self->priv->prog);

	/* Restore texture environment */
	glPopAttrib();

	glDisable(GL_POINT_SPRITE);

	if(self->priv->texture) lw_texture_disable(self->priv->texture);
}

static void
noise_particle_system_init(NoiseParticleSystem *self)
{
    self->priv = NOISE_PARTICLE_SYSTEM_GET_PRIVATE(self);

	self->priv->particle_count = 0;
	self->priv->fading_time = 1000;

	self->priv->datadir = NULL;

    self->priv->particles = g_array_new(FALSE, TRUE, sizeof(Particle));
    self->priv->vertices = NULL;
    self->priv->alpha = NULL;
    self->priv->size = NULL;

    /* Create buffers */
    glGenBuffers(1, &self->priv->vertex_buffer);
	glGenBuffers(1, &self->priv->alpha_buffer);
	glGenBuffers(1, &self->priv->size_buffer);

	self->priv->prog = NULL;
	self->priv->texture = NULL;
}

static void
noise_particle_system_dispose(GObject *object)
{
	NoiseParticleSystem *self = NOISE_PARTICLE_SYSTEM(object);

	if(self->priv->prog)
	{
		g_object_unref(self->priv->prog);
		self->priv->prog = NULL;
	}

	if(self->priv->texture)
	{
		g_object_unref(self->priv->texture);
		self->priv->texture = NULL;
	}

	/* Chain up to the parent class */
	G_OBJECT_CLASS(noise_particle_system_parent_class)->dispose(object);
}

static void
noise_particle_system_finalize(GObject *object)
{
	NoiseParticleSystem *self = NOISE_PARTICLE_SYSTEM(object);

	g_free(self->priv->datadir);

	g_array_free(self->priv->particles, TRUE);
	g_free(self->priv->vertices);
	g_free(self->priv->alpha);
	g_free(self->priv->size);

	if(self->priv->vertex_buffer)
		glDeleteBuffers(1, &self->priv->vertex_buffer);

	if(self->priv->alpha_buffer)
		glDeleteBuffers(1, &self->priv->alpha_buffer);

	if(self->priv->size_buffer)
		glDeleteBuffers(1, &self->priv->size_buffer);

	/* Chain up to the parent class */
	G_OBJECT_CLASS(noise_particle_system_parent_class)->finalize(object);
}

static void
noise_particle_system_class_init(NoiseParticleSystemClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->set_property = noise_particle_system_set_property;
	gobject_class->get_property = noise_particle_system_get_property;
	gobject_class->dispose = noise_particle_system_dispose;
	gobject_class->finalize = noise_particle_system_finalize;

	g_type_class_add_private(klass, sizeof(NoiseParticleSystemPrivate));

	obj_properties[PROP_PARTICLE_COUNT] =
		g_param_spec_uint("particle-count",
		                  "Number of particles",
		                  "Number of particles",
		                  0,
		                  300,
		                  0,
		                  G_PARAM_READWRITE);

	obj_properties[PROP_FADING_TIME] =
		g_param_spec_uint("fading-time",
		                  "Fading time",
		                  "The time a particle needs to appear or disappear",
		                  0,
		                  5000,
		                  1000,
		                  G_PARAM_READWRITE);

	g_object_class_install_properties(gobject_class, N_PROPERTIES, obj_properties);
}

